/*
 * Copyright (C) 2015 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.android.tools.idea.npw.assetstudio.ui;

import com.android.assetstudiolib.GraphicGenerator;
import com.android.resources.Density;
import com.android.tools.idea.npw.assetstudio.icon.CategoryIconMap;
import com.android.tools.idea.ui.ImageComponent;
import com.google.common.collect.Maps;
import com.google.common.collect.Ordering;
import com.google.common.primitives.Ints;
import com.intellij.openapi.ui.VerticalFlowLayout;
import com.intellij.ui.JBColor;
import com.intellij.ui.components.JBLabel;
import com.intellij.util.IconUtil;
import com.intellij.util.ui.JBImageIcon;
import com.intellij.util.ui.UIUtil;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import javax.swing.*;
import javax.swing.border.LineBorder;
import java.awt.*;
import java.awt.image.BufferedImage;
import java.util.Collection;
import java.util.Map;

/**
 * A panel which shows the icons generated by a {@link GraphicGenerator}.
 */
@SuppressWarnings("UseJBColor")
public final class PreviewIconsPanel extends JPanel {

  private final Theme myTheme;
  private final Map<Density, ImageComponent> myIconImages = Maps.newHashMap();

  @Nullable private CategoryIconMap.Filter myFilter;

  private JPanel myRootPanel;
  private JPanel myIconsPanel;
  private JBLabel myTitleLabel;

  public PreviewIconsPanel(@NotNull String title, @NotNull Theme theme) {
    super(new BorderLayout());
    add(myRootPanel);

    myTitleLabel.setText(title);
    myTitleLabel.setVisible(!title.isEmpty());

    myTheme = theme;
    myRootPanel.setBackground(myTheme.getMainColor());
    myRootPanel.setOpaque(myTheme != Theme.TRANSPARENT);
    myTitleLabel.setForeground(myTheme.getAltColor());
  }

  public PreviewIconsPanel(@NotNull String title, @NotNull Theme theme, @NotNull CategoryIconMap.Filter filter) {
    this(title, theme);
    myFilter = filter;
  }

  private static void updateImage(@NotNull ImageComponent imageComponent, @NotNull BufferedImage sourceImage) {
    JBImageIcon icon = IconUtil.createImageIcon(sourceImage);
    Dimension d = new Dimension(icon.getIconWidth(), icon.getIconHeight());
    imageComponent.setPreferredSize(d);
    imageComponent.setMinimumSize(d);
    imageComponent.setIcon(icon);
  }

  /**
   * Update the preview images in this panel using values generated by a {@link GraphicGenerator}.
   *
   * Note: This panel lazily instantiates its components based on the values in the map the first
   * time this is called. Additional calls should pass in maps with the same keys. This happens
   * automatically as long as you keep generating icons with consistent options.
   */
  public void updateImages(@NotNull CategoryIconMap categoryIconMap) {

    Map<Density, BufferedImage> densityImageMap;
    if (myFilter != null) {
      densityImageMap = categoryIconMap.toDensityMap(myFilter);
    }
    else {
      densityImageMap = categoryIconMap.toDensityMap();
    }

    if (myIconsPanel.getComponentCount() == 0) {
      initializeIconComponents(densityImageMap.keySet());
    }

    for (Map.Entry<Density, BufferedImage> densityImageEntry : densityImageMap.entrySet()) {
      updateImage(myIconImages.get(densityImageEntry.getKey()), densityImageEntry.getValue());
    }
  }

  private void initializeIconComponents(@NotNull Collection<Density> densities) {
    // Sort the densities, so that we always line up icons from highest to lowest densities.
    Ordering<Density> densityOrdering = new Ordering<Density>() {
      @Override
      public int compare(Density left, Density right) {
        // The lower the enum index, the higher the density
        return Ints.compare(left.ordinal(), right.ordinal());
      }
    };

    densities = densityOrdering.immutableSortedCopy(densities);
    for (Density density : densities) {
      JPanel iconPanel = new JPanel(new VerticalFlowLayout(false, false));
      iconPanel.setBackground(myTheme.getMainColor());
      iconPanel.setOpaque(myTheme != Theme.TRANSPARENT);

      JBLabel title = new JBLabel(density.getResourceValue());
      title.setForeground(myTheme.getAltColor());
      iconPanel.add(title);

      ImageComponent iconImage = new ImageComponent(null);
      iconImage.setBorder(new LineBorder(myTheme.getAltColor()));
      iconImage.setOpaque(false);
      iconPanel.add(iconImage);

      GridBagConstraints c = new GridBagConstraints();
      c.anchor = GridBagConstraints.PAGE_START;

      myIconsPanel.add(iconPanel, c);
      myIconImages.put(density, iconImage);
    }
  }

  /**
   * Color themes which are useful for acting as backdrops to our various icon types.
   */
  public enum Theme {
    DARK(Color.BLACK, Color.WHITE),
    LIGHT(Color.WHITE, Color.BLACK),
    GRAY(Color.DARK_GRAY, Color.LIGHT_GRAY),
    TRANSPARENT(UIUtil.TRANSPARENT_COLOR, JBColor.BLACK);

    private final Color myMainColor;
    private final Color myAltColor;

    Theme(@NotNull Color mainColor, @NotNull Color altColor) {
      myMainColor = mainColor;
      myAltColor = altColor;
    }

    /**
     * The theme's main color, used for the background.
     */
    @NotNull
    public Color getMainColor() {
      return myMainColor;
    }

    /**
     * The theme's alt color, used for rendering lines and text on the background.
     */
    @NotNull
    public Color getAltColor() {
      return myAltColor;
    }
  }
}
